/*
    (C) 1995-97 AROS - The Amiga Research OS
    $Id: tasklist.c,v 1.6 2019/08/12 20:45:10 piru Exp $

    Desc:
    Lang: english
*/

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <exec/tasks.h>
#include <exec/system.h>
#include <exec/execbase.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <aros/machine.h>
#include <strings.h>
#include "stackdump.h"
#include "tasklist_version.h"

static const char version[] = VERSTAG;

#define	MAXNAMESIZE	256
#define MAXLEVEL	30
#define PPCSTACKSCANSIZE	8192

struct task
{
	struct task *next;
	char name[MAXNAMESIZE+1];
	char cliname[MAXNAMESIZE+1];

	ULONG pid;
	APTR address;
	WORD type;
	WORD state;
	void *spreg;
	void *spupper;
	void *splower;
	ULONG stacksize;
	ULONG stackused;

	ULONG ppcstacksize;
	ULONG ppcstackused;
	void *ppcspupper;
	void *ppcsplower;
	UBYTE *srr0;
	UBYTE *lr;
	UBYTE *ctr;
	ULONG cr;
	ULONG xer;
	ULONG gpr[32];
	double fpr[32];
#warning "Needs alignment..but works for now because the os uses memcpy instead of vector instructions"
	ULONG vpr[4*32];
	ULONG vscr[4];
	ULONG altivec;
	ULONG vsave;
	WORD pri;
	ULONG sigwait;
	ULONG sigrecvd;
	ULONG sigexcept;
	struct SegEntry SegArray[0];
};

#define ARG_TEMPLATE "NAME,ADDRESS/N,VERBOSE/S,STACKTRACE/S,STACKLEVEL/N,NORUN/S,NOWAIT/S,NOREADY/S,INTERNAL/S,REGDUMP/S,REGCHECK/S"

enum
{
	ARG_NAME = 0,
	ARG_ADDRESS,
	ARG_VERBOSE,
	ARG_STACKTRACE,
	ARG_STACKLEVEL,
	ARG_NORUN,
	ARG_NOWAIT,
	ARG_NOREADY,
	ARG_INTERNAL,
	ARG_REGDUMP,
	ARG_REGCHECK,
	NOOFARGS
};


void *memclr(void *, size_t);

static int fillbuffer(struct task **buffer, IPTR size, struct ExecBase *SysBase);


struct ExecBase	*SysBase;
struct Library  *SysDebugBase;
struct DosLibrary *DOSBase;
UBYTE		*Emulation;
ULONG		EmulationSize;
UBYTE		*Module;
ULONG		ModuleSize;
UBYTE		*TaskExitCode;
UBYTE		*TaskExitCode_M68k;
ULONG		MaxLevel;
ULONG		InternalFlag;

IPTR		args[NOOFARGS] = { (IPTR)0 };

int main(void)
{
	struct RDArgs *rda;
	IPTR size;
	struct task *buffer,*tasks,*tasks2;
	LONG error=RETURN_FAIL;

	SysBase=*((struct ExecBase**) 4);
	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37)))
	{
		if ((SysDebugBase=OpenLibrary("sysdebug.library",0)))
		{
			rda=(struct RDArgs *)ReadArgs(ARG_TEMPLATE, args, NULL);

			if (rda != NULL)
			{
				error=RETURN_OK;

				MaxLevel = args[ARG_STACKLEVEL] ? *((ULONG*)args[ARG_STACKLEVEL]) : MAXLEVEL;

				InternalFlag = args[ARG_INTERNAL];

				Emulation = NULL;
				EmulationSize = 0;
				Module = NULL;
				ModuleSize = 0;

				NewGetSystemAttrs(&Emulation,
				                  sizeof(Emulation),
				                  SYSTEMINFOTYPE_EMULATION_START,
				                  TAG_DONE);

				NewGetSystemAttrs(&EmulationSize,
				                  sizeof(EmulationSize),
				                  SYSTEMINFOTYPE_EMULATION_SIZE,
				                  TAG_DONE);

				NewGetSystemAttrs(&Module,
				                  sizeof(Module),
				                  SYSTEMINFOTYPE_MODULE_START,
				                  TAG_DONE);

				NewGetSystemAttrs(&ModuleSize,
				                  sizeof(ModuleSize),
				                  SYSTEMINFOTYPE_MODULE_SIZE,
				                  TAG_DONE);


				NewGetSystemAttrs(&TaskExitCode,
				                  sizeof(TaskExitCode),
				                  SYSTEMINFOTYPE_TASKEXITCODE,
				                  TAG_DONE);

				NewGetSystemAttrs(&TaskExitCode_M68k,
				                  sizeof(TaskExitCode_M68k),
				                  SYSTEMINFOTYPE_TASKEXITCODE_M68K,
				                  TAG_DONE);

                                //Printf("TaskExitCode 0x%lx\n",TaskExitCode);


				for(size=65536*2;;size*=2)
				{
					buffer=AllocVec(size,MEMF_ANY);
					tasks=buffer;
					if(buffer==NULL)
					{
						error = RETURN_FAIL;
						break;
					}
					if (fillbuffer(&tasks,size,SysBase))
					{
						FPuts(Output(),"   pid    address type  pri   state 68kstack/used     ppcstack/used     name\n");
						for(tasks2=buffer;tasks2<tasks;tasks2=tasks2->next)
						{
							char *state;
							char *type;
							ULONG cli=FALSE;
							UBYTE stackusedbuf[11];
							UBYTE *stackused;

							if (tasks2->type==NT_TASK)
							{
								type = "task";
							}
							else
							if (tasks2->type==NT_PROCESS)
							{
								type = "proc";
							}
							else
							{
								cli = TRUE;
								type = " cli";
							}

							if (tasks2->state==TS_RUN)
							{
								state = "    run";
							}
							else
							if (tasks2->state==TS_READY)
							{
								state = "  ready";
							}
							else
							if (tasks2->state==TS_WAIT)
							{
								if (tasks2->sigwait==0)
								{
									state = "   dead";
								}
								else
								{
									state = "   wait";
								}
							}
							else
							if (tasks2->state==TS_EXCEPT)
							{
								state = " except";
							}
							else
							if (tasks2->state==TS_REMOVED)
							{
								state = "removed";
							}
							else
							{
								state = "-------";
							}

							if (tasks2->stackused != ~0)
							{
								NewRawDoFmt("%-8lu", NULL, stackusedbuf, tasks2->stackused);
								stackused = stackusedbuf;
							}
							else
							{
								stackused = "???     ";
							}
							if (cli)
							{
								Printf("%6lu 0x%08lx %s %4ld %s %8lu/%s %8lu/%-8lu %s [%s]\n",
									tasks2->pid,
									(ULONG)tasks2->address,
									(ULONG)type,
									(LONG)tasks2->pri,
									(ULONG)state,
									tasks2->stacksize,
									(ULONG)stackused,
									tasks2->ppcstacksize,
									tasks2->ppcstackused,
									(ULONG)tasks2->name,
									(ULONG)tasks2->cliname);
							}
							else
							{
								Printf("%4lu 0x%08lx %s %4ld %s %8lu/%s %8lu/%-8lu %s\n",
									tasks2->pid,
									(ULONG)tasks2->address,
									(ULONG)type,
									(LONG)tasks2->pri,
									(ULONG)state,
									tasks2->stacksize,
									(ULONG)stackused,
									tasks2->ppcstacksize,
									tasks2->ppcstackused,
									(ULONG)tasks2->name);
							}

							if (args[ARG_VERBOSE])
							{
								Printf(":     State %ld\n",
									(LONG)tasks2->state);
								Printf(":   SigWait 0x%08lx\n",
									tasks2->sigwait);
								Printf(": SigExcept 0x%08lx\n",
									tasks2->sigexcept);
								Printf(":  SigRecvd 0x%08lx\n",
									tasks2->sigrecvd);

								Printf(":M68k SPUpper 0x%08lx\n",
									(ULONG)tasks2->spupper);

								Printf(":M68k SPLower 0x%08lx\n",
									(ULONG)tasks2->splower);

								Printf(":M68k  SPReg 0x%08lx\n",
									(ULONG)tasks2->spreg);

								Printf(": PPC SPUpper 0x%08lx\n",
									(ULONG)tasks2->ppcspupper);

								Printf(": PPC SPLower 0x%08lx\n",
									(ULONG)tasks2->ppcsplower);
							}

							if (args[ARG_STACKTRACE])
							{
								if(tasks2->state!=TS_RUN)
								{
									int i;


									ShowPtr("                  SRR0",tasks2->srr0);
									ShowPtr("                    LR",tasks2->lr);
									ShowPtr("                   CTR",tasks2->ctr);

									for (i=0;tasks2->SegArray[i].Address && i<MaxLevel;i++)
									{
										if (tasks2->SegArray[i].Name)
										{
											Printf("     StackFrame[%2ld].LR -> Address 0x%08lx -> %s%s%s Hunk %ld Offset 0x%08lx\n",
												i,
												(ULONG)tasks2->SegArray[i].Address,
												(ULONG)tasks2->SegArray[i].Name,
												(ULONG)(tasks2->SegArray[i].Function ? "/" : ""),
												(ULONG)(tasks2->SegArray[i].Function ? (const char *)tasks2->SegArray[i].Function : ""),
												tasks2->SegArray[i].Segment,
												tasks2->SegArray[i].Offset);
										}
										else
										{
											if (((UBYTE*) tasks2->SegArray[i].Address == TaskExitCode))
											{
												Printf("     StackFrame[%2ld].LR -> Address 0x%08lx -> ABOX: Module <exec.library/TaskExitCode>\n",
													i,
													(ULONG)tasks2->SegArray[i].Address);
											}
											else
											if (((UBYTE*) tasks2->SegArray[i].Address == TaskExitCode_M68k))
											{
												Printf("     StackFrame[%2ld].LR -> Address 0x%08lx -> ABOX: Module <exec.library/TaskExitCode_M68k>\n",
													i,
													(ULONG)tasks2->SegArray[i].Address);
											}
											else
											if (((UBYTE*) tasks2->SegArray[i].Address >= Emulation) && ((UBYTE*) tasks2->SegArray[i].Address < (Emulation + EmulationSize)))
											{
												Printf("     StackFrame[%2ld].LR -> Address 0x%08lx -> ABOX: Emulation Offset 0x%lx\n",
													i,
													(ULONG)tasks2->SegArray[i].Address,
													(UBYTE*) tasks2->SegArray[i].Address - Emulation);
											}
											else
											if (((UBYTE*) tasks2->SegArray[i].Address >= Module) && ((UBYTE*) tasks2->SegArray[i].Address < (Module + ModuleSize)))
											{
												Printf("     StackFrame[%2ld].LR -> Address 0x%08lx -> ABOX: Module Offset 0x%lx\n",
													i,
													(ULONG)tasks2->SegArray[i].Address,
													(UBYTE*) tasks2->SegArray[i].Address - Module);

											}
											else
											{
												Printf("     StackFrame[%2ld].LR -> Address 0x%08lx\n",
													i,
													(ULONG)tasks2->SegArray[i].Address);
											}

										}

									}

								}
							}

							if (args[ARG_REGDUMP])
							{
								int i;

								if(tasks2->state!=TS_RUN)
								{
									Printf("     SRR0 0x%08lx LR 0x%08lx CTR 0x%lx CR 0x%lx XER 0x%lx\n",
										(ULONG)tasks2->srr0,
										(ULONG)tasks2->lr,
										(ULONG)tasks2->ctr,
										tasks2->cr,
										tasks2->xer);

									for (i=0;i<4;i++)
									{
										Printf("     GPR[%02ld] 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",
											i*8,
											tasks2->gpr[i*8+0],
											tasks2->gpr[i*8+1],
											tasks2->gpr[i*8+2],
											tasks2->gpr[i*8+3],
											tasks2->gpr[i*8+4],
											tasks2->gpr[i*8+5],
											tasks2->gpr[i*8+6],
											tasks2->gpr[i*8+7]);
									}
#if 0
									for (i=0;i<4;i++)
									{
										Printf("     FPR[%02ld] %g %g %g %g %g %g %g %g\n",
											i*8,
											tasks2->fpr[i*8+0],
											tasks2->fpr[i*8+1],
											tasks2->fpr[i*8+2],
											tasks2->fpr[i*8+3],
											tasks2->fpr[i*8+4],
											tasks2->fpr[i*8+5],
											tasks2->fpr[i*8+6],
											tasks2->fpr[i*8+7]);
									}
#endif
									if (tasks2->altivec)
									{
										Printf("     VSAVE 0x%08lx VSCR 0x%08lx 0x%08lx 0x%08lx 0x%08lx \n",
											(ULONG)tasks2->vsave,
											(ULONG)tasks2->vscr[0],
											(ULONG)tasks2->vscr[1],
											(ULONG)tasks2->vscr[2],
											(ULONG)tasks2->vscr[3]);

										for (i=0;i<32;i++)
										{
											Printf("     VPR[%02ld] 0x%08lx 0x%08lx 0x%08lx 0x%08lx | 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n",
												i*2,
												tasks2->vpr[i*2+0],
												tasks2->vpr[i*2+1]);
										}
									}
								}
							}

							if (args[ARG_REGCHECK])
							{
								int i;

								if(tasks2->state!=TS_RUN)
								{
									for (i=0;i<32;i++)
									{
										ShowReg(i,(UBYTE*) tasks2->gpr[i]);
									}
								}
							}

							if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
							{
								error = RETURN_FAIL;
								SetIoErr(ERROR_BREAK);
								break;
							}
						}
						FreeVec(buffer);
						break;
					}
					FreeVec(buffer);
				}
				FreeArgs(rda);
			}
			if (error != RETURN_OK)
			{
				PrintFault(IoErr(), NULL);
			}
			CloseLibrary(SysDebugBase);
		}
		CloseLibrary((struct Library *)DOSBase);
	}

	return error;
}

static int addtask(struct Task *task, struct task **t, STRPTR bufend)
{
	STRPTR s;
	STRPTR tptr;
	STRPTR name;
	ULONG type, state;
	ULONG len;
	ULONG tmp;
	int i;

	if (!NewGetTaskAttrs(task,
	                     &type,
	                     sizeof(type),
	                     TASKINFOTYPE_TYPE,
	                     TAG_DONE))
	{
		type = task->tc_Node.ln_Type;
	}

	if (args[ARG_NAME])
	{
		if (strcmp((char*) args[ARG_NAME],
			   task->tc_Node.ln_Name))
		{
			if(type==NT_PROCESS && ((struct Process *)task)->pr_CLI)
			{
				/* Use cli_CommandName field for the name - Piru
				 */

				s = BADDR(((struct CommandLineInterface *) BADDR(((struct Process *)task)->pr_CLI))->cli_CommandName);
				if (s && s[0])
				{
					if (strlen((char*) args[ARG_NAME]) == s[0])
					{
						if (memcmp((char*) args[ARG_NAME],s + 1,s[0])==0)
						{
						}
						else
						{
							return 1;
						}
					}
					else
					{
						return 1;
					}
				}
				else
				{
					return 1;
				}
			}
			else
			{
				return 1;
			}
		}

	}

	if (args[ARG_ADDRESS])
	{
		if (((struct Task*) *((ULONG*) (args[ARG_ADDRESS]))) != task)
		{
			return 1;
		}
	}

	if ((STRPTR) (*t + 1) > bufend)
		return 0;

	if ((*t)->name)
	{
		strncpy((*t)->name,
			task->tc_Node.ln_Name,
			MAXNAMESIZE-1);
		(*t)->name[MAXNAMESIZE] = '\0';
	}
	else
	{
		(*t)->name[0] = '\0';
	}

	if (!NewGetTaskAttrs(task,
	                     &((*t)->pid),
	                     sizeof((*t)->pid),
	                     TASKINFOTYPE_PID_CLI,
	                     TAG_DONE))
	{
		(*t)->pid = 0;
	}
	(*t)->address=task;

	/* type, pri, state */

	(*t)->type=type;

	if (!NewGetTaskAttrs(task,
	                     &tmp,
	                     sizeof(tmp),
	                     TASKINFOTYPE_PRI,
	                     TAG_DONE))
	  tmp = (WORD)task->tc_Node.ln_Pri;
	(*t)->pri = (WORD)tmp;

	if (!NewGetTaskAttrs(task,
	                     &state,
	                     sizeof(state),
	                     TASKINFOTYPE_STATE,
	                     TAG_DONE))
	  state = task->tc_State;
	(*t)->state = state;


	/* 68k stack */

	/*if (!NewGetTaskAttrs(task,
	                     &((*t)->spreg),
	                     sizeof((*t)->spreg),
	                     TASKINFOTYPE_SPREG_M68K,
	                     TAG_DONE))*/
	{
		if(state==TS_RUN)
		{
			(*t)->spreg=(APTR)REG_A7;
		}
		else
		{
			(*t)->spreg=task->tc_SPReg;
		}
	}

	if (!NewGetTaskAttrs(task,
	                     &((*t)->stacksize),
	                     sizeof((*t)->stacksize),
	                     TASKINFOTYPE_STACKSIZE_M68K,
	                     TAG_DONE))
		(*t)->stacksize=(ULONG) task->tc_SPUpper - (ULONG) task->tc_SPLower;

	(*t)->stackused = ~0; /* default to "unknown" */
	if (!NewGetTaskAttrs(task,
	                     &((*t)->stackused),
	                     sizeof((*t)->stackused),
	                     TASKINFOTYPE_USEDSTACKSIZE_M68K,
	                     TAG_DONE) ||
	                     (state==TS_RUN &&
	                      (SysBase->LibNode.lib_Version < 50 ||
	                       (SysBase->LibNode.lib_Version == 50 && 
	                        SysBase->LibNode.lib_Revision <= 69))))
	{
		(*t)->stackused=(ULONG) task->tc_SPUpper - (ULONG) (*t)->spreg;
	}

	if (!NewGetTaskAttrs(task,
	                     &((*t)->spupper),
	                     sizeof((*t)->spupper),
	                     TASKINFOTYPE_SPUPPER_M68K,
	                     TAG_DONE))
		(*t)->spupper=task->tc_SPUpper;

	if (!NewGetTaskAttrs(task,
	                     &((*t)->splower),
	                     sizeof((*t)->splower),
	                    TASKINFOTYPE_SPLOWER_M68K,
	                    TAG_DONE))
		(*t)->splower=task->tc_SPLower;


	/* ppc stack */

	NewGetTaskAttrs(task,
	                &((*t)->ppcstacksize),
	                sizeof((*t)->ppcstacksize),
	                TASKINFOTYPE_STACKSIZE,
	                TAG_DONE);

	NewGetTaskAttrs(task,
	                &((*t)->ppcspupper),
	                sizeof((*t)->ppcspupper),
	                TASKINFOTYPE_SPUPPER,
	                TAG_DONE);

	NewGetTaskAttrs(task,
	                &((*t)->ppcsplower),
	                sizeof((*t)->ppcsplower),
	                TASKINFOTYPE_SPLOWER,
	                TAG_DONE);

	NewGetTaskAttrs(task,
	                &((*t)->ppcstackused),
	                sizeof((*t)->ppcstackused),
	                TASKINFOTYPE_USEDSTACKSIZE,
	                TAG_DONE);

	if (state==TS_RUN &&
	    (SysBase->LibNode.lib_Version < 50 ||
	    (SysBase->LibNode.lib_Version == 50 && 
             SysBase->LibNode.lib_Revision <= 69)))
	{
		register ULONG REG_GPR1 __asm("r1");
		(*t)->ppcstackused = (ULONG) (*t)->ppcspupper - REG_GPR1;
	}


	/* signals */

	if (!NewGetTaskAttrs(task,
	                     &((*t)->sigwait),
	                     sizeof((*t)->sigwait),
	                     TASKINFOTYPE_SIGWAIT,
	                     TAG_DONE))
		(*t)->sigwait=task->tc_SigWait;

	if (!NewGetTaskAttrs(task,
	                     &((*t)->sigrecvd),
	                     sizeof((*t)->sigrecvd),
	                     TASKINFOTYPE_SIGRECVD,
	                     TAG_DONE))
		(*t)->sigrecvd=task->tc_SigRecvd;

	if (!NewGetTaskAttrs(task,
	                     &((*t)->sigexcept),
	                     sizeof((*t)->sigexcept),
	                     TASKINFOTYPE_SIGEXCEPT,
	                     TAG_DONE))
		(*t)->sigexcept=task->tc_SigExcept;


	(*t)->cliname[0] = '\0';

	if(type==NT_PROCESS && ((struct Process *)task)->pr_CLI)
	{
		/* Use cli_CommandName field for the name - Piru
		*/

		s = BADDR(((struct CommandLineInterface *) BADDR(((struct Process *)task)->pr_CLI))->cli_CommandName);
		if (s && s[0])
		{
			strncpy((*t)->cliname,
				s + 1,
				s[0]);
			(*t)->cliname[s[0]] = '\0';
		}

		(*t)->type=-1;
	}

	tptr = (STRPTR) (*t + 1);

	if(state!=TS_RUN)
	{

		NewGetTaskAttrs(task,
		                &((*t)->srr0),
		                sizeof((*t)->srr0),
		                TASKINFOTYPE_PPC_SRR0,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                &((*t)->lr),
		                sizeof((*t)->lr),
		                TASKINFOTYPE_PPC_LR,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                &((*t)->ctr),
		                sizeof((*t)->ctr),
		                TASKINFOTYPE_PPC_CTR,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                &((*t)->cr),
		                sizeof((*t)->cr),
		                TASKINFOTYPE_PPC_CR,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                &((*t)->xer),
		                sizeof((*t)->xer),
		                TASKINFOTYPE_PPC_XER,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                ((*t)->gpr),
		                sizeof((*t)->gpr),
		                TASKINFOTYPE_PPC_GPR,
		                TASKINFOTAG_REGSTART,0,
		                TASKINFOTAG_REGCOUNT,32,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                ((*t)->fpr),
		                sizeof((*t)->fpr),
		                TASKINFOTYPE_PPC_FPR,
		                TASKINFOTAG_REGSTART,0,
		                TASKINFOTAG_REGCOUNT,32,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                &((*t)->vsave),
		                sizeof((*t)->vsave),
		                TASKINFOTYPE_PPC_VSAVE,
		                TAG_DONE);

		NewGetTaskAttrs(task,
		                ((*t)->vscr),
		                sizeof((*t)->vscr),
		                TASKINFOTYPE_PPC_VSCR,
		                TAG_DONE);

#if 1
		(*t)->altivec = 0;
#else
		(*t)->altivec = NewGetTaskAttrs(task,
		 	        		((*t)->vpr),
			        	        sizeof((*t)->vpr),
			                	TASKINFOTYPE_PPC_VMX,
		        		        TASKINFOTAG_REGSTART,0,
				                TASKINFOTAG_REGCOUNT,32,
				                TAG_DONE);
#endif
	}

	if (args[ARG_STACKTRACE])
	{

		if ((STRPTR) (&((*t)->SegArray[MaxLevel+1])) > bufend)
		{
			return(0);
		}

		memclr(&(*t)->SegArray[0],(MaxLevel+1) * sizeof(struct SegEntry));

		if (((*t)->gpr[1] >= (ULONG) (*t)->ppcsplower) &&
		    ((*t)->gpr[1] < (ULONG) (*t)->ppcspupper))
		{
			ShowPPCStackHistory((ULONG*) (*t)->gpr[1],
					   (ULONG*) (*t)->ppcspupper,
					   (*t)->SegArray,
					   MaxLevel);
		}
		else
		{
			ShowPPCStackHistory((ULONG*) (*t)->gpr[1],
					   (ULONG*) ((*t)->gpr[1] + PPCSTACKSCANSIZE),
					   (*t)->SegArray,
					   MaxLevel);
		}

		for (i=0;i<MaxLevel;i++)
		{
			if (!((*t)->SegArray[i].Address))
			{
				break;
			}
		}

		tptr += (i+1) * sizeof(struct SegEntry);

		for (i=0;i<MaxLevel;i++)
		{
			if ((*t)->SegArray[i].Address)
			{
				if ((name=(*t)->SegArray[i].Name))
				{
					len = strlen(name);
					if ((tptr + len + 1) > bufend)
					{
						return(0);
					}
					strcpy(tptr,
						name);

					(*t)->SegArray[i].Name	=	tptr;
					tptr += len + 1;

					if ((name=(*t)->SegArray[i].Function))
					{
						len = strlen(name);
						if ((tptr + len + 1) > bufend)
						{
							return(0);
						}
						strcpy(tptr,
							name);

						(*t)->SegArray[i].Function	=	tptr;
						tptr += len + 1;
					}
				}
			}
			else
			{
				break;
			}
		}
	}
	tptr = (STRPTR) (((IPTR) tptr + 3) & ~0x3UL);

	if(tptr > bufend)
		return 0;

	(*t)->next = (struct task*) tptr;

	(*t) = (struct task*) tptr;
	return 1;
}

static int fillbuffer(struct task **buffer, IPTR size, struct ExecBase *SysBase)
{
	STRPTR bufend = ((STRPTR)*buffer)+size;
	struct Task *task;
	Disable();
	if (!args[ARG_NORUN])
	{
		if(!addtask(SysBase->ThisTask,buffer,bufend))
		{
			Enable();
			return 0;
		}
	}
	if (!args[ARG_NOREADY])
	{
		for(task=(struct Task *)SysBase->TaskReady.lh_Head;
			task->tc_Node.ln_Succ!=NULL;
			task=(struct Task *)task->tc_Node.ln_Succ)
		{
			if(!addtask(task,buffer,bufend))
			{
				Enable();
				return 0;
			}
		}
	}
	if (!args[ARG_NOWAIT])
	{
		for(task=(struct Task *)SysBase->TaskWait.lh_Head;
			task->tc_Node.ln_Succ!=NULL;
			task=(struct Task *)task->tc_Node.ln_Succ)
		{
			if(!addtask(task,buffer,bufend))
			{
				Enable();
				return 0;
			}
		}
	}
	Enable();
	return 1;
}

